# Copyright 2004, @Last Software, Inc., modified by D. Bur

# This software is provided as an example of using the Ruby interface
# to SketchUp.

# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name        :   extrudelinetool
# Description :   A script to extrude lines along a vector given with 2 clicks
# Menu Item   :   Plugins->Extrude lines (vector by 2 points)
# Usage       :   Select lines, select "Extrude lines (vector by 2 points)" from the plugins Menu,
#                 then click 2 points on your model
# Date        :   10/18/2004
# Type        :   tool
#-----------------------------------------------------------------------------
### TIG tweaked so faces orient and smoothness and curves kept etc...

require 'sketchup.rb'

module DBUR

class ExtrudeLineTool###

class Sketchup::Face
 def orient_connected_faces
    @connected_faces=[]
		self.all_connected.each{|e|
			if e.class==Sketchup::Face
			e.edges.each{|edge|
				if edge.faces[1]
					@connected_faces << e
					break
				end#if
			}
			end#if
		}
		@connected_faces=[self] + @connected_faces 
		@connected_faces.uniq!
		@awaiting_faces=@connected_faces
    @processed_faces=[self]
    @done_faces=[]
    msg="Orienting Faces"
    ###
    timeout=1000
    timer=0
		while @awaiting_faces[0]
      msg=msg+"."
			@processed_faces.each{|face|
					if ! @done_faces.include?(face)
					Sketchup::set_status_text(msg,SB_PROMPT)
				@face=face
						face_flip
					end#if
			}
      timer+=1
      @awaiting_faces=[] if timer=timeout
    end#while
		Sketchup::set_status_text((""),SB_PROMPT)
 end#def
 def face_flip
    @awaiting_faces=@awaiting_faces-[@face]
    @face.edges.each{|edge|
      rev1=edge.reversed_in?(@face)
      @common_faces=edge.faces-[@face]
      @common_faces.each{|face|
				rev2=edge.reversed_in?(face)
				face.reverse! if @awaiting_faces.include?(face) && rev1==rev2
				@awaiting_faces=@awaiting_faces-[face]
				@processed_faces<<face
			}
    }
    @done_faces<<@face
 end#def
end#class

def getExtents
	bbox=Sketchup.active_model.bounds
	bbox.add(@ip.position)if @ip && @ip.valid?
	bbox.add(@ip1.position)if @ip1 && @ip1.valid?
	bbox.add(@ip2.position)if @ip2 && @ip2.valid?
	return bbox
end

def initialize
	@ss = Sketchup.active_model.selection
	@ip1 = nil
	@ip2 = nil
	@xdown = 0
	@ydown = 0
end

def activate
	@ip1 = Sketchup::InputPoint.new
	@ip2 = Sketchup::InputPoint.new
	@ip = Sketchup::InputPoint.new
	@drawn = false
	self.reset(nil)
end

def deactivate(view)
  view.invalidate if @drawn
end

def onMouseMove(flags, x, y, view)
	if( @state == 0 )
		@ip.pick(view, x, y)
		if( @ip != @ip1 )
			view.invalidate if( @ip.display? || @ip1.display? )
			@ip1.copy! @ip
			view.tooltip = @ip1.tooltip
		end
	else
		@ip2.pick(view, x, y, @ip1)
		view.tooltip = @ip2.tooltip if( @ip2.valid? )
		view.invalidate
		if( @ip2.valid? )
			length = @ip1.position.distance(@ip2.position)
			Sketchup::set_status_text length.to_s, SB_VCB_VALUE
		end
		if( (x-@xdown).abs > 10 || (y-@ydown).abs > 10 )
			@dragging = true
		end
	end
end

def onLButtonDown(flags, x, y, view)
	if( @state == 0 )
		@ip1.pick(view, x, y)
		if( @ip1.valid? )
			@state = 1
			Sketchup::set_status_text "拉线成面:点击终点", SB_PROMPT
			@xdown = x
			@ydown = y
		end
	else
		if( @ip2.valid? )
			self.create_geometry(@ip1.position, @ip2.position,view)
			self.reset(view)
		end
	end
	view.lock_inference
end

# The onLButtonUp method is called when the user releases the left mouse button.
def onLButtonUp(flags, x, y, view)
	if( @dragging && @ip2.valid? )
		self.create_geometry(@ip1.position, @ip2.position,view)
		self.reset(view)
	end
end

def onKeyDown(key, repeat, flags, view)
	if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
		@shift_down_time = Time.now
		if( view.inference_locked? )
			view.lock_inference
		elsif( @state == 0 && @ip1.valid? )
			view.lock_inference @ip1
		elsif( @state == 1 && @ip2.valid? )
			view.lock_inference @ip2, @ip1
		end
	end
end

def onKeyUp(key, repeat, flags, view)
	if( key == CONSTRAIN_MODIFIER_KEY && view.inference_locked? && (Time.now - @shift_down_time) > 0.5 )
		view.lock_inference
	end
end

def onUserText(text, view)
	return if ! @state == 1
	return if ! @ip2.valid?
	begin
		value = text.to_l
	rescue
		# Error parsing the text
		puts "拉线成面: \n\n无法将 #{text} 转换为长度。"
		value = nil
		Sketchup::set_status_text "", SB_VCB_VALUE
	end
	return if ! value
	pt1 = @ip1.position
	vec = @ip2.position - pt1
	if( vec.length == 0.0 )
			return
	end
	vec.length = value
	#pt2 = pt1 + vec
	pt2=pt1.offset vec
	#puts (pt1.distance pt2).to_s
	self.create_geometry(pt1, pt2, view)
	self.reset(view)
end


def draw(view)
	if( @ip1.valid? )
		if( @ip1.display? )
			@ip1.draw(view)
			@drawn = true
			view.draw_points(@ip1.position, 8, 1, "black")
		end
		if( @ip2.valid? )
			@ip2.draw(view) if( @ip2.display? )	
			view.set_color_from_line(@ip1, @ip2)
			self.draw_geometry(@ip1.position, @ip2.position, view)
			@drawn = true
		end
	end
	view.draw_points(@ip2.position, 8, 1, "black")
end

# onCancel is called when the user hits the escape key
def onCancel(flag, view)
  self.reset(view)
end


# Reset the tool back to its initial state
def reset(view)
	@state = 0
	
	Sketchup::set_status_text("拉线成面:点击起点", SB_PROMPT)
	
	@ip1.clear
	@ip2.clear
	
	if( view )
		view.tooltip = nil
		view.invalidate if @drawn
	end
	
	@drawn = false
	@dragging = false
end

# Create new geometry when the user has selected two points.
def create_geometry(p1, p2, view)

	model = Sketchup.active_model
	model.start_operation("拉线成面", true)
	entities = model.active_entities
	ss = model.selection
	indice_line = 0

	if ss.empty? 
		UI.messagebox("拉线成面: \n\n没有选择线，请先选择线之后再进行相关操作！")
		return nil
	end

	# Selection error checking
	###others = 0
	###i = 0
	###0.upto(ss.length - 1) do |something|
	###  element = ss[i]
	###  if( element.typename != "Edge")
	###    others = others + 1
	###  end
	###  i = i + 1
	###end  #of upto

	###if( others != 0 )
		###UI.messagebox( "Extrude Lines: \n\n"+ others.to_s + " Selected Entities are NOT Lines.\n[they will be ignored]")
	##end

	dx = p2.x - p1.x
	dy = p2.y - p1.y
	dz = p2.z - p1.z

	sel=model.selection.to_a
	ssa=[]
	model.selection.each{|e| ssa<< e if e.class == Sketchup::Edge}
	model.selection.clear
	### ssa is now just the edges

	cverts=[]
	edges=[]
	ssa.each{|e|
		cverts<< e.curve.vertices if e.curve && ! cverts.include?(e.curve.vertices)
		edges<< e if ! e.curve
	}
	cpoints=[]
	cverts.each{|verts|
		pts=[]
		verts.each{|vert|pts<<vert.position}
		cpoints<< pts
	}
	group=model.active_entities.add_group()
	gp=group.entities.add_group()
	edges.each{|e|gp.entities.add_line(e.start.position,e.end.position)}
	cpoints.each{|pts|gp.entities.add_curve(pts)}
	ssa=gp.entities.to_a
	gp_top=gp.copy
	tr=Geom::Transformation.translation(p1.vector_to(p2))
	gp_top.entities.transform_entities(tr, gp_top.entities.to_a)
	cpoints_top=[]
	cpoints.each{|points|
		pts=[]
		points.each{|point|
			pt=point.clone
			pt.transform!(tr)
			pts << pt
		}
		cpoints_top<<pts
	}
	clones=gp_top.entities.to_a
	cverts=[]
	clones.each{|e|cverts<<[e.vertices[0].position,e.vertices[1].position]}
	ssa=gp.explode
	clones=gp_top.explode
	entities=group.entities

	0.upto(ssa.length-1) do |i|
		edge=ssa[i]
		next unless edge && edge.is_a?(Sketchup::Edge) && edge.valid?
	 begin
		v1 = edge.vertices[0].position
		v2 = edge.vertices[1].position
		v1_top = Geom::Point3d.new(v1.x+dx, v1.y+dy, v1.z+dz)
		v2_top = Geom::Point3d.new(v2.x+dx, v2.y+dy, v2.z+dz)
		e1=entities.add_line(v1, v1_top)
		e2=entities.add_line(v2_top, v2)
		if edge.curve || (edge.smooth? || edge.soft?)
			mid=0
			e1.start.edges.each{|e|mid=mid+1 if e.curve}
			e2.start.edges.each{|e|mid=mid+1 if e.curve}
			if mid==4
				e1.smooth = true; e1.soft = true
				e2.smooth = true; e2.soft = true
			end#if
		end#if
		e1.find_faces
	 rescue
		###
	 end#begin
		i = i + 1 ### next edge
	end # of upto
	###
	face=nil
	group.entities.each{|e|
		if e.class==Sketchup::Face
			face=e
			break
		end#if
	}
	face.orient_connected_faces if face ### orient faces
	###
	### reform curves
	sel.each{|e|e.explode_curve if e.class==Sketchup::Edge && e.curve}
	group.entities.each{|e|e.explode_curve if e.class==Sketchup::Edge}
	group.explode
	###
	gpx=model.active_entities.add_group()
	cpoints.each{|pts|cvs=gpx.entities.add_curve(pts)}
	cpoints_top.each{|pts|cvs_top=gpx.entities.add_curve(pts)}
	gpx.explode
	### allow for 'merged' edges in selection...
	clones.grep(Sketchup::Edge).each{|e|model.selection.add(e) if e.valid?}
	cverts.each{|verts|
		model.active_entities.to_a.each{|e|
			if e.valid? && e.class==Sketchup::Edge
				if (e.vertices[0].position==verts[0] || e.vertices[0].position==verts[1]) && (e.vertices[1].position==verts[0] || e.vertices[1].position==verts[1])
					model.selection.add(e)### as it's a matching edge
				end#if
			end#if
		}
	}

end#def

# Draw the geometry
def draw_geometry(pt1, pt2, view)
  view.draw_line(pt1, pt2)
end

end # class ExtrudeLineTool

#-----------------------------------------------------------------------------
# This functions is just a shortcut for selecting the new tool
#def ex_linetool
    #Sketchup.active_model.select_tool ExtrudeLineTool.new
#end

if( not file_loaded?("extrudelinetool.rb") )
    add_separator_to_menu("Plugins")
    UI.menu("Plugins").add_item("拉线成面 [TIG'd]"){ Sketchup.active_model.select_tool(ExtrudeLineTool.new()) }
end
#-----------------------------------------------------------------------------
file_loaded("extrudelinetool.rb")

    # 以下部分是添加图标的代码
    toolbar = UI::Toolbar.new "拉线成面"
    cmd = UI::Command.new("拉线成面") { Sketchup.active_model.select_tool(ExtrudeLineTool.new()) }
    cmd.tooltip = "拉线成面"
    cmd.status_bar_text = '将线挤出为面（需要先选择线）。'
    cmd.small_icon = "extrudelinetool-Modified-TIG/Extrude Line Too icon.png"
    cmd.large_icon = "extrudelinetool-Modified-TIG/Extrude Line Too icon.png"
    toolbar = toolbar.add_item cmd
    toolbar.show
    puts cmd.large_icon
    # 以上部分是添加图标的代码    

end#module